import socket
import getpass
from pathlib import Path

import grpc
from ..common.grpc import grpc_pb2
from ..common.grpc import grpc_pb2_grpc
from ..common.utils import *

# wlab credentials
server_ip = '164.67.195.207'
server_port = '61005'

options = [
      ('grpc.max_send_message_length', 100 * 1024 * 1024),
      ('grpc.max_receive_message_length', 100 * 1024 * 1024),
]

# Server.crt
certs_path = Path(__file__).resolve().parent.parent/'core'/'certs'/'server.crt'
with certs_path.open('rb') as f:
    trusted_certs = f.read()
    
credentials = grpc.ssl_channel_credentials(root_certificates=trusted_certs)
channel = grpc.secure_channel(f'{server_ip}:{server_port}', credentials, options=options)
stub = grpc_pb2_grpc.GenericRPCStub(channel)

tcp_calls = 0

def get_tcp_calls():
    return tcp_calls
        
def rpc_client(*, function_name, args):
    global tcp_calls
    tcp_calls += 1
    # print(tcp_calls)
    # if not is_connected:
    #     response = rpc_client(function_name="UserLogin", args={"username": grpc_pb2.Argument(string_value=input("Username: ")), "password": grpc_pb2.Argument(string_value=getpass.getpass("Password: ")), "client_ip": grpc_pb2.Argument(string_value=local_ip)})
    #     if (response.results['status'].string_value == 'Success'):
    #         print("Login successful.")
    #         is_connected = True
    
    # TODO: Handle user login
    
    # print(f"Opening connection to {server_ip}:{server_port}")
    
    # TODO: Handle Errors
    
    # print(f"Calling function: {function_name}")
    response = stub.Call(grpc_pb2.GenericRPCRequest(function_name=function_name, args=args))
    
    if 'a' in response.results:
        print(f"Error: {unmap_arg(response.results['a'])}")
        exit()
        
    if 'UE' in response.results:
        print(f"UserError: {unmap_arg(response.results['UE'])}")
        input("Hit enter to continue...")
            
    
    return response

#region Example Usage

# if __name__ == '__main__':
#     args={
#         "key1": grpc_pb2.Argument(string_value="Hello"),
#         "key2": grpc_pb2.Argument(int32_value=123),
#         "key3": grpc_pb2.Argument(float_value=4.56),
#         "key4": grpc_pb2.Argument(bool_value=True),
#         "client_ip": grpc_pb2.Argument(string_value=local_ip)
#     }
    
#     response = rpc_client(function_name="echo", args=args)
    
#     if 'client_ip' in response.results:
#         del response.results['client_ip']
    
#     # Print results
#     print("Received response:")
#     for key, arg in response.results.items():
#         # Decode the oneof fields
#         if arg.HasField('string_value'):
#             value = arg.string_value
#         elif arg.HasField('int32_value'):
#             value = arg.int32_value
#         elif arg.HasField('float_value'):
#             value = arg.float_value
#         elif arg.HasField('bool_value'):
#             value = arg.bool_value
#         else:
#             value = "Undefined"
            
#         print(f"{key}: {value}")

#endregion
    



from ...core.grpc_client import rpc_client
from ...common.utils import *
from ...common.grpc import grpc_pb2
from ...core.grpc_client import get_tcp_calls

def try_get(function_name, token):
    try:
        return unmap_arg(rpc_client(function_name=f"Pluto:{function_name}:GET", args={'a':map_arg(token)}).results[function_name])
    except Exception as e:
        input(f"Error: {e}\nHit enter to continue...")
    return None

def try_set(function_name, value, token):
    try:
        rpc_client(function_name=f"Pluto:{function_name}:SET", args={function_name: map_arg(value), 'a':map_arg(token)})
    except Exception as e:
        input(f"Error: {e}\nHit enter to continue...")
        
def try_call_0_arg(function_name, token):   # 0 argument call
    try:
        response = rpc_client(
            function_name=f"Pluto:{function_name}:CALL0", 
            args={
                'a': map_arg(token)
            }
        )
        return unmap_arg(response.results[function_name])
    except Exception as e:
        input(f"RPC_0_call Error: {e}\nHit enter to continue...")
    return None
        
def try_call_1_arg(function_name, arg, token):  # 1 argument call
    try:
        response = rpc_client(
            function_name=f"Pluto:{function_name}:CALL1", 
            args={
                'a': map_arg(token),
                'arg1': map_arg(arg)
            }
        )
        # The server should return something like {function_name: <something>}
        return unmap_arg(response.results[function_name])
    except Exception as e:
        input(f"RPC_1_call Error: {e}\nHit enter to continue...")
    return None

class rx_def:
    pass

class tx_def:
    pass

class rx_tx_def(rx_def, tx_def):
    pass

class ad9364(rx_tx_def):
    pass

class Pluto: # client
    
    def __init__(self, token='', debug=False):
        self.token = token
        # self.try_set(function_name="ip", value=grpc_pb2.Argument(string_value=ip))
        
        if debug:
            print("Pluto Remote Client Initialized.")
        
    def api_token(self, token:str) -> None:
        self.token = token
    
    # PlutoSDR
    
    _device_name = "PlutoSDR"
    _uri_auto = "ip:pluto.local"

    def __repr__(self): # ! UNTESTED !
        return try_get("__repr__", self.token)
    
    #region ad9364
    """AD9364 Transceiver"""

    @property
    def filter(self):
        return try_get("filter", self.token)

    @filter.setter
    def filter(self, value):
        try_set("filter", value, self.token)

    @property
    def loopback(self):
        """loopback: Set loopback mode. Options are:
        0 (Disable), 1 (Digital), 2 (RF)"""
        return try_get("loopback", self.token)

    @loopback.setter
    def loopback(self, value):
        try_set("loopback", value, self.token)

    @property
    def gain_control_mode_chan0(self):
        """gain_control_mode_chan0: Mode of receive path AGC. Options are:
        slow_attack, fast_attack, manual"""
        return try_get("gain_control_mode_chan0", self.token)

    @gain_control_mode_chan0.setter
    def gain_control_mode_chan0(self, value):
        try_set("gain_control_mode_chan0", value, self.token)

    @property
    def rx_hardwaregain_chan0(self):
        """rx_hardwaregain_chan0: Gain applied to RX path. Only applicable when
        gain_control_mode is set to 'manual'"""
        return try_get("rx_hardwaregain_chan0", self.token)

    @rx_hardwaregain_chan0.setter
    def rx_hardwaregain_chan0(self, value):
        try_set("rx_hardwaregain_chan0", value, self.token)

    @property
    def tx_hardwaregain_chan0(self):
        """tx_hardwaregain_chan0: Attenuation applied to TX path"""
        return try_get("tx_hardwaregain_chan0", self.token)

    @tx_hardwaregain_chan0.setter
    def tx_hardwaregain_chan0(self, value):
        try_set("tx_hardwaregain_chan0", value, self.token)

    @property
    def rx_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX path"""
        return try_get("rx_rf_bandwidth", self.token)

    @rx_rf_bandwidth.setter
    def rx_rf_bandwidth(self, value):
        try_set("rx_rf_bandwidth", value, self.token)

    @property
    def tx_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX path"""
        return try_get("tx_rf_bandwidth", self.token)

    @tx_rf_bandwidth.setter
    def tx_rf_bandwidth(self, value):
        try_set("tx_rf_bandwidth", value, self.token)

    @property
    def sample_rate(self):                # ! UNTESTED !
        """sample_rate: Sample rate RX and TX paths in samples per second"""
        return try_get("sample_rate", self.token)

    @sample_rate.setter
    def sample_rate(self, rate):          # ! UNTESTED !
        try_set("sample_rate", rate, self.token)

    @property
    def rx_lo(self):
        """rx_lo: Carrier frequency of RX path"""
        return try_get("rx_lo", self.token)

    @rx_lo.setter
    def rx_lo(self, value):
        try_set("rx_lo", value, self.token)

    @property
    def tx_lo(self):
        """tx_lo: Carrier frequency of TX path"""
        return try_get("tx_lo", self.token)

    @tx_lo.setter
    def tx_lo(self, value):
        try_set("tx_lo", value, self.token)
        
    @property
    def tx_cyclic_buffer(self):
        """tx_cyclic_buffer: Size of cyclic buffer"""
        return try_get("tx_cyclic_buffer", self.token)
    
    @tx_cyclic_buffer.setter
    def tx_cyclic_buffer(self, value):
        try_set("tx_cyclic_buffer", value, self.token)
        
    def tx_destroy_buffer(self):
        try_call_0_arg("tx_destroy_buffer", self.token)

    def rx_destroy_buffer(self):
        try_call_0_arg("rx_destroy_buffer", self.token)

    #endregion
    
    #region rx_def
    
    def rx(self):
        return try_get("rx", self.token)
    
    @property
    def rx_buffer_size(self):
        return try_get("rx_buffer_size", self.token)
    
    @rx_buffer_size.setter
    def rx_buffer_size(self, value):
        try_set("rx_buffer_size", value, self.token)
    
    #endregion
    
    #region tx_def
    
    def tx(self, value):
        return try_call_1_arg("tx", value, self.token)
    
    # @tx.setter
    # def tx(self, value):
    #     try_set("tx", value, self.token)
    
    #endregion
    
    #region tx
    
    #endregion
    
    #region _dec_int_fpga_filter
    
    """Decimator and interpolator fpga filter controls"""

    def _get_rates(self, dev, output):           # ! UNTESTED !
        """Get the decimation and interpolation rates"""
        return try_get("rates", self.token)

    @property
    def rx_dec8_filter_en(self) -> bool:         # ! UNTESTED !
        """rx_dec8_filter_en: Enable decimate by 8 filter in FPGA"""
        return try_get("rx_dec8_filter_en", self.token)

    @rx_dec8_filter_en.setter
    def rx_dec8_filter_en(self, value: bool):    # ! UNTESTED !
        """rx_dec8_filter_en: Enable decimate by 8 filter in FPGA"""
        return try_set("rx_dec8_filter_en", value, self.token)

    @property
    def tx_int8_filter_en(self) -> bool:         # ! UNTESTED !
        """tx_int8_filter_en: Enable interpolate by 8 filter in FPGA"""
        return try_get("tx_int8_filter_en", self.token)

    @tx_int8_filter_en.setter
    def tx_int8_filter_en(self, value: bool):    # ! UNTESTED !
        """tx_int8_filter_en: Enable interpolate by 8 filter in FPGA"""
        return try_set("tx_int8_filter_en", value, self.token)
    
    #endregion
 


from .grpc_client import rpc_client
from ..common.utils import *

import datetime

class RemoteRFAccount:
    def __init__(self, username:str=None, password:str=None, email:str=None):
        self.username = username
        self.password = password
        self.email = email
    
    def create_user(self):
        response = rpc_client(function_name="ACC:create_user", args={"un":map_arg(self.username), "pw":map_arg(self.password), "em":map_arg(self.email)})
        if 'UC' in response.results:
            print(f'User {unmap_arg(response.results["UC"])} successfully created.')
            return True
        elif 'UE' in response.results:
            print(f'Error: {unmap_arg(response.results["UE"])}')
            return False
    
    def login_user(self):
        username = self.username
        password = self.password
        response = rpc_client(function_name="ACC:login", args={"un":map_arg(username), "pw":map_arg(password)})
        if 'UC' in response.results:
            print(f'User {unmap_arg(response.results["UC"])} successful login.')
            return True
        elif 'UE' in response.results:
            print(f'Error: {unmap_arg(response.results["UE"])}')
            return False
    
    def reserve_device(self, device_id:int, start_time:datetime, end_time:datetime):
        response = rpc_client(function_name="ACC:reserve_device", args={"un":map_arg(self.username), "pw":map_arg(self.password), "dd":map_arg(device_id), "st":map_arg(int(start_time.timestamp())), "et":map_arg(int(end_time.timestamp()))})
        
        if 'ace' in response.results:
            raise Exception(f'{unmap_arg(response.results["ace"])}')
        elif 'Token' in response.results:
            return unmap_arg(response.results["Token"])
            
    def get_reservations(self):
        return rpc_client(function_name='ACC:get_res', args={"un":map_arg(self.username), "pw":map_arg(self.password)})
    
    def get_devices(self):
        return rpc_client(function_name='ACC:get_dev', args={"un":map_arg(self.username), "pw":map_arg(self.password)})
    
    def cancel_reservation(self, res_id:int):
        return rpc_client(function_name='ACC:cancel_res', args={"un":map_arg(self.username), "pw":map_arg(self.password), "res_id":map_arg(res_id)})
    
    def get_perms(self):
        return rpc_client(function_name='ACC:get_perms', args={"un":map_arg(self.username), "pw":map_arg(self.password)})
    
    



from setuptools import setup, find_packages
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="remoteRF",
    version="0.0.7.21",
    author="Ethan Ge",
    author_email="ethoGalaxy@gmail.com",
    description="A python API to remotely access signal centric hardware. Client-side only!",
    long_description=long_description,  # Set the README content here
    long_description_content_type="text/markdown",  # Specify that it's Markdown
    packages=find_packages(where="src"),  # Automatically finds subpackages like core, deviceA, deviceB
    package_dir={"": "src"},
    license_file='MIT',
    include_package_data=True,  # Includes files specified in MANIFEST.in
    install_requires=[
        "grpcio", "protobuf", "numpy", "prompt_toolkit", "python-dotenv"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    entry_points={
        'console_scripts': [
            'remoterf-login=remoteRF.core.acc_login:main',
        ],
    },
)